@@ -1,22 +1,25 @@
 """Lower level API, configuration, and HTTP stuff."""
 
-import json
-import urlparse
+from ubersmith.compat import total_ordering, file_type
 import time
-from functools import partial
 
-import httplib2
+import requests
 
 from ubersmith.exceptions import (
     RequestError,
     ResponseError,
     UpdatingTokenResponse,
+    MaintenanceResponse,
+)
+from ubersmith.utils import (
+    append_qs,
+    to_nested_php_args,
+    get_filename,
 )
-from ubersmith.utils import append_qs, urlencode_unicode, to_nested_php_args
 
 __all__ = [
     'METHODS',
-    'HttpRequestHandler',
+    'RequestHandler',
     'get_default_request_handler',
     'set_default_request_handler',
 ]
@@ -29,18 +32,15 @@ METHODS = {
     u'client.ach_delete': u'Delete a Bank Account',
     u'client.ach_update': u'Update a Bank Account',
     u'client.add': u'Add a New Client',
-    u'client.avatar_get': u"Retrieve a Client Avatar",
-    u'client.avatar_set': u"Set a Client Avatar",
     u'client.cc_add': u'Add a New Credit Card',
     u'client.cc_delete': u'Delete a Credit Card',
+    u'client.cc_info': u"List a Client's Credit Card Details",
     u'client.cc_update': u'Update a Credit Card',
     u'client.comment_list': u"List a Client's Comments",
     u'client.contact_add': u'Add a New Contact',
     u'client.contact_delete': u'Delete a Contact',
     u'client.contact_get': u'Get Contact Details',
     u'client.contact_list': u"List a Client's Contacts",
-    u'client.contact_metadata_get': u"Get a Contact's Metadata",
-    u'client.contact_metadata_single': u"Get a Contact's Metadata Value",
     u'client.contact_update': u'Update a Contact',
     u'client.count': u'Count Active Clients',
     u'client.credit_add': u'Add an Account Credit',
@@ -84,10 +84,6 @@ METHODS = {
     u'client.service_prorate': u'Prorate a Service',
     u'client.service_update': u'Update a Service',
     u'client.set_login': u"Set a Client's Login",
-    u'client.tax_exemption_add': u"Add a new Tax Exemption",
-    u'client.tax_exemption_get': u"Get a Client's Tax Exemption",
-    u'client.tax_exemption_list': u"List a Client's Tax Exemptions",
-    u'client.tax_exemption_update': u"Update a Client's Tax Exemption",
     u'client.update': u'Update a Client',
     u'device.add': u'Add a New Device',
     u'device.comment_list': u"List a Device's Comments",
@@ -102,7 +98,6 @@ METHODS = {
     u'device.ip_assignment_delete': u'Delete a Device IP Assignment',
     u'device.ip_assignment_list': u'List Device IP Assignments',
     u'device.ip_assignment_update': u'Update a Device IP Assignment',
-    u'device.ip_block_list': u"List IP Blocks",
     u'device.ip_get_available': u'List Available IP Addresses',
     u'device.ip_get_unassigned': u'Get Unassigned IP Addresses',
     u'device.ip_group_add': u'Add a Device IP Group',
@@ -110,7 +105,6 @@ METHODS = {
     u'device.ip_group_list': u'List a Device IP Group',
     u'device.ip_group_update': u'Update a Device IP Group',
     u'device.ip_lookup': u'Look Up a Device IP',
-    u'device.ip_pool_list': u"List IP Pools",
     u'device.ip_unassign': u'Unassign a Device IP',
     u'device.list': u'List Devices',
     u'device.module_call': u'Call a Device Module Function',
@@ -157,17 +151,11 @@ METHODS = {
     u'support.ticket_post_staff_response': u'Post a Staff Response to a Ticket',
     u'support.ticket_submit': u'Submit a New Ticket',
     u'support.ticket_submit_outgoing': u'Create a New Outgoing Ticket',
-    u'support.ticket_type_list': u"Get a List of Ticket Types",
     u'support.ticket_update': u'Update a Ticket',
-    u'uber.admin_avatar_get': u"Retrieve an Admin Avatar",
-    u'uber.admin_avatar_set': u"Set an Admin Avatar",
-    u'uber.admin_get': u"User Information",
-    u'uber.admin_list': u"List User Logins",
     u'uber.api_export': u'Export Data',
     u'uber.attachment_get': u'Get an attachment',
     u'uber.attachment_list': u'List Attachments',
     u'uber.check_login': u'Verify a login and password',
-    u'uber.client_permission_list': u"List available permissions",
     u'uber.client_welcome_stats': u'Display Client Statistics',
     u'uber.comment_add': u'Add Comment',
     u'uber.comment_delete': u'Delete Comment',
@@ -176,11 +164,6 @@ METHODS = {
     u'uber.comment_update': u'Update Comment',
     u'uber.documentation': u'Download API Documentation',
     u'uber.event_list': u'Access the Event Log',
-    u'uber.file_add': u"Add a file",
-    u'uber.file_delete': u"Delete a file",
-    u'uber.file_get': u"Get a File",
-    u'uber.file_list': u"Get a List of Files",
-    u'uber.file_update': u"Update a file",
     u'uber.forgot_pass': u'Send a Password Reminder',
     u'uber.login_list': u'List User Logins',
     u'uber.mail_get': u'Get an Email From the Log',
@@ -194,8 +177,6 @@ METHODS = {
     u'uber.quick_stats_detail': u'Get Detailed System Stats',
     u'uber.service_plan_get': u'Get Service Plan Details',
     u'uber.service_plan_list': u'List Service Plans',
-    u'uber.tax_exemption_type_get': u"Get a Tax Exemption Type",
-    u'uber.tax_exemption_type_list': u"List Tax Exemption Types",
     u'uber.user_exists': u'Check whether a Client Exists',
     u'uber.username_exists': u'Check Whether a Username Exists',
 }
@@ -210,68 +191,99 @@ class _ProxyModule(object):
         """Return the call with request_handler prefilled."""
         call_func = getattr(self.module, name)
         if callable(call_func):
-            call_p = partial(call_func, request_handler=self.handler)
+            call_p = call_func.handler(self.handler)
             # store partial on proxy so it doesn't have to be created again
             setattr(self, name, call_p)
             return call_p
         raise AttributeError("'{0}' object has no attribute '{1}'".format(
-                                                   type(self).__name__, name))
+            type(self).__name__, name))
 
 
-class _AbstractRequestHandler(object):
-    def process_request(self, method, data=None, raw=False):
-        """Process request.
+class RequestHandler(object):
+    """Handles HTTP requests and authentication."""
+
+    def __init__(self, base_url, username=None, password=None, verify=True):
+        """Initialize HTTP request handler with optional authentication.
+
+            base_url: URL to send API requests
+            username: Username for API access
+            password: Password for API access
+            verify: Verify HTTPS certificate
+
+        """
+        self.base_url = base_url
+        self.username = username
+        self.password = password
+        self.verify = verify
+
+    def process_request(self, method, data=None):
+        """Process request over HTTP to ubersmith instance.
 
             method: Ubersmith API method string
             data: dict of method arguments
-            raw: Set to True to return the raw response vs the default
-                 behavior of returning JSON data
 
         """
-        raise NotImplementedError
+        # make sure requested method is valid
+        self._validate_request_method(method)
 
-    def _render_response(self, response, content, raw):
-        """Render response as python object.
+        # attempt the request three times
+        attempts = range(3)
+        for i in attempts:
+            response = self._send_request(method, data)
 
-            response: dict like object with headers
-            content: raw response string from ubersmith
-            raw: Set to True to return the raw response vs the default
-                 behavior of returning JSON data
-
-        """
-        # just return the raw response
-        if raw:
-            return response, content
-
-        # response isn't json
-        if response.get('content-type') != 'application/json':
             # handle case where ubersmith is 'updating token'
             # see: https://github.com/jasonkeene/python-ubersmith/issues/1
-            if response.get('content-type') == 'text/html' and \
-                'Updating Token' in content:
-                raise UpdatingTokenResponse
-            raise ResponseError("Response wasn't application/json")
+            if all([
+                response.headers.get('content-type') == 'text/html',
+                'Updating Token' in response.content,
+            ]):
+                if i != attempts[-1]:
+                    # wait 2 secs before retrying request
+                    time.sleep(2)
+                    continue
+                else:
+                    raise UpdatingTokenResponse
+            break
 
-        # response is json
-        response_dict = json.loads(content)
+        resp = BaseResponse(response)
 
         # test for error in json response
-        if not response_dict.get('status'):
-            raise ResponseError(response=response_dict)
+        if response.headers.get('content-type') == 'application/json':
+            if not resp.json.get('status'):
+                if all([
+                    resp.json.get('error_code') == 1,
+                    resp.json.get('error_message') == u"We are currently "
+                        "undergoing maintenance, please check back shortly.",
+                ]):
+                    raise MaintenanceResponse(response=resp.json)
+                else:
+                    raise ResponseError(response=resp.json)
+        return resp
 
-        # make sure to handle condition where empty string is sent for data :/
-        return response_dict['data'] or {}
+    def _send_request(self, method, data):
+        url = append_qs(self.base_url, {'method': method})
+        data, files, headers = self._encode_data(data)
+        return requests.post(url, data=data, files=files, headers=headers,
+                             auth=(self.username, self.password),
+                             verify=self.verify)
 
-    def _validate_request_method(self, method):
+    @staticmethod
+    def _validate_request_method(method):
         """Make sure requested method is valid."""
         if method not in METHODS:
             raise RequestError("Requested method is not valid.")
 
-    def _encode_data(self, data):
+    @staticmethod
+    def _encode_data(data):
         """URL encode data."""
         data = data if data is not None else {}
         data = to_nested_php_args(data)
-        return urlencode_unicode(data)
+        files = dict([
+            (key, value) for key, value in
+            data.items() if isinstance(value, file_type)])
+        for fname in files:
+            del data[fname]
+        return data, files or None, None
 
     def __getattr__(self, name):
         """If attribute accessed is a call module, return a proxy."""
@@ -283,75 +295,218 @@ class _AbstractRequestHandler(object):
             setattr(self, name, proxy)
             return proxy
         raise AttributeError("'{0}' object has no attribute '{1}'".format(
-                                                   type(self).__name__, name))
+            type(self).__name__, name))
 
 
-class HttpRequestHandler(_AbstractRequestHandler):
-    """Handles HTTP requests and authentication."""
+class BaseResponse(object):
+    """Wraps response object and emulates different types."""
+    def __init__(self, response):
+        self.response = response  # requests' response object
 
-    def __init__(self, base_url, username=None, password=None):
-        """Initialize HTTP request handler with optional authentication.
+    @classmethod
+    def from_cleaned(cls, response, cleaned):
+        resp = cls(response.response)
+        resp.cleaned = cleaned
+        return resp
 
-            base_url: URL to send API requests
-            username: Username for API access
-            password: Password for API access
+    @property
+    def json(self):
+        return self.response.json()
 
-        >>> handler = HttpRequestHandler('http://127.0.0.1:8088/')
-        >>> handler.base_url
-        'http://127.0.0.1:8088/'
-        >>> config = {
-        ...     'base_url': 'http://127.0.0.1/api/',
-        ...     'username': 'admin',
-        ...     'password': 'test_pass',
-        ... }
-        >>> handler = HttpRequestHandler(**config)
-        >>> handler.base_url
-        'http://127.0.0.1/api/'
-        >>> handler.username
-        'admin'
-        >>> handler.password
-        'test_pass'
+    @property
+    def data(self):
+        if hasattr(self, "cleaned"):
+            return self.cleaned
+        else:
+            return self.json['data']
 
-        """
-        self.base_url = base_url
-        self.username = username
-        self.password = password
+    @property
+    def type(self):
+        return self.response.headers.get('content-type')
 
-        self._http = httplib2.Http(disable_ssl_certificate_validation=True)
-        self._http.add_credentials(self.username, self.password,
-                                   urlparse.urlparse(self.base_url)[1])
 
-    def process_request(self, method, data=None, raw=False):
-        """Process request over HTTP to ubersmith instance.
+@total_ordering
+class DictResponse(BaseResponse):
+    def keys(self):
+        return self.data.keys()
 
-            method: Ubersmith API method string
-            data: dict of method arguments
-            raw: Set to True to return the raw response vs the default
-                 behavior of returning JSON data
+    def values(self):
+        return self.data.values()
 
-        """
-        # make sure requested method is valid
-        self._validate_request_method(method)
+    def items(self):
+        return self.data.items()
 
-        # try request 3 times
-        for i in range(3):
-            # make the request
-            response, content = self._send_request(method, data)
-            try:
-                # render the response as python object
-                return self._render_response(response, content, raw)
-            except UpdatingTokenResponse:
-                # wait 4 secs before retrying request
-                time.sleep(4)
-        # if last attempt still threw an exception, reraise it
-        raise
+    def get(self, key, default=None):
+        return self.data.get(key, default)
 
-    def _send_request(self, method, data):
-        url = append_qs(self.base_url, {'method': method})
-        body = self._encode_data(data)
-        # httplib2 requires that you manually send Content-Type on POSTs :/
-        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
-        return self._http.request(url, "POST", body, headers)
+    def __iter__(self):
+        return iter(self.data)
+
+    def __getitem__(self, key):
+        return self.data[key]
+
+    def __len__(self):
+        return len(self.data)
+
+    def __eq__(self, other):
+        return self.data == other
+
+    def __lt__(self, other):
+        return self.data < other
+
+    def __contains__(self, item):
+        return item in self.data
+
+
+@total_ordering
+class IntResponse(BaseResponse):
+    @property
+    def numerator(self):
+        return self.data
+
+    @property
+    def denominator(self):
+        return 1
+
+    @property
+    def real(self):
+        return self.data
+
+    @property
+    def imag(self):
+        return 0
+
+    def bit_length(self):
+        if hasattr(self.data, 'bit_length'):
+            return self.data.bit_length()
+        else:
+            return len(bin(abs(self.data))) - 2
+
+    def conjugate(self):
+        return self.data
+
+    def __int__(self):
+        return self.data
+    __index__ = __long__ = __trunc__ = __int__
+
+    def __float__(self):
+        return float(self.data)
+
+    def __oct__(self):
+        return oct(self.data)
+
+    def __hex__(self):
+        return hex(self.data)
+
+    def __eq__(self, other):
+        return self.data == other
+
+    def __lt__(self, other):
+        return self.data < other
+
+    def __add__(self, other):
+        return int(self) + other
+    __radd__ = __add__
+
+    def __sub__(self, other):
+        return int(self) - other
+
+    def __rsub__(self, other):
+        return other - int(self)
+
+    def __mul__(self, other):
+        return int(self) * other
+    __rmul__ = __mul__
+
+    def __div__(self, other):
+        return int(self) / other
+
+    def __rdiv__(self, other):
+        return other / int(self)
+
+    def __floordiv__(self, other):
+        return int(self) // other
+
+    def __rfloordiv__(self, other):
+        return other // int(self)
+
+    def __truediv__(self, other):
+        return float(self) / other
+
+    def __rtruediv__(self, other):
+        return other / float(self)
+
+    def __mod__(self, other):
+        return int(self) % other
+
+    def __rmod__(self, other):
+        return other % int(self)
+
+    def __pow__(self, other):
+        return int(self) ** other
+
+    def __rpow__(self, other):
+        return other ** int(self)
+
+    def __abs__(self):
+        return abs(self.data)
+
+    def __neg__(self):
+        return -self.data
+
+    def __pos__(self):
+        return self.data
+
+    def __divmod__(self, other):
+        return self // other, self % other
+
+    def __rdivmod__(self, other):
+        return other // self, other % self
+
+    def __and__(self, other):
+        return self.data & other
+    __rand__ = __and__
+
+    def __or__(self, other):
+        return self.data | other
+    __ror__ = __or__
+
+    def __xor__(self, other):
+        return self.data ^ other
+    __rxor__ = __xor__
+
+    def __lshift__(self, other):
+        return self.data << other
+
+    def __rlshift__(self, other):
+        return other << self.data
+
+    def __rshift__(self, other):
+        return self.data >> other
+
+    def __rrshift__(self, other):
+        return other >> self.data
+
+    def __invert__(self):
+        return ~self.data
+
+    def __nonzero__(self):
+        return bool(self.data)
+
+
+class FileResponse(BaseResponse):
+    @property
+    def json(self):
+        raise NotImplementedError
+
+    @property
+    def data(self):
+        return self.response.content
+
+    @property
+    def filename(self):
+        disposition = self.response.headers.get('content-disposition')
+        return get_filename(disposition)
 
 
 def get_default_request_handler():
@@ -363,7 +518,7 @@ def get_default_request_handler():
 
 def set_default_request_handler(request_handler):
     """Set the default request handler."""
-    if not isinstance(request_handler, _AbstractRequestHandler):
+    if not isinstance(request_handler, RequestHandler):
         raise TypeError(
             "Attempted to set an invalid request handler as default.")
     global _DEFAULT_REQUEST_HANDLER
