@@ -6,19 +6,13 @@ ubersmith.calls.BaseCall.
 
 """
 
-from ubersmith.calls import GroupCall
+# from ubersmith.calls import BaseCall, GroupCall
 from ubersmith.utils import prepend_base
 
-__all__ = [
-    'DepartmentListCall',
-]
+__all__ = []
 
 _ = prepend_base(__name__.split('.')[-1])
 
 
-class DepartmentListCall(GroupCall):
-    method = _('department_list')
-    int_fields = [
-        'q_id',
-        'q_order',
-    ]
+# class (BaseCall):
+#     method = _('')
