@@ -1,17 +1,29 @@
 """Utility functions used throughout ubersmith library."""
 import inspect
-import urllib
-import urlparse
+try:
+    from collections.abc import MutableSequence, MutableMapping
+except ImportError:  # pragma: no cover
+    from collections import MutableSequence, MutableMapping
+try:
+    from urllib import parse as urlparse
+except ImportError:  # pragma: no cover
+    import urlparse
+try:
+    from urllib.parse import urlencode
+except ImportError:  # pragma: no cover
+    from urllib import urlencode
+
+from six import string_types, text_type
 
 __all__ = [
     'append_qs',
     'urlencode_unicode',
+    'to_nested_php_args',
     'prepend_base',
     'isdict',
     'islist',
     'isstr',
-    'signature_position',
-    'to_nested_php_args',
+    'get_filename'
 ]
 
 
@@ -24,11 +36,6 @@ def append_qs(url, query_string):
         * a dict of lists of strings: {'test3': ['val1', 'val2']}
         * a list of tuples: [('test3', 'val1'), ('test3', 'val2')]
 
-    >>> url = 'http://domain.tld/path/?test1=val&test2#hash'
-    >>> qs = {'test3': 'val'}
-    >>> append_qs(url, qs)
-    'http://domain.tld/path/?test1=val&test2=&test3=val#hash'
-
     """
     parsed_url = urlparse.urlsplit(url)
     parsed_qs = urlparse.parse_qsl(parsed_url.query, True)
@@ -45,7 +52,7 @@ def append_qs(url, query_string):
     elif islist(query_string):
         parsed_qs += query_string
     else:
-        raise ValueError('Unexpected query_string value')
+        raise TypeError('Unexpected query_string type')
 
     return urlparse.urlunsplit((
         parsed_url.scheme,
@@ -56,19 +63,17 @@ def append_qs(url, query_string):
     ))
 
 
-# TODO this should match the same input as urllib.urlencode so it can be used
-# as a drop in replacement
-def urlencode_unicode(data):
+def urlencode_unicode(data, doseq=0):
     """urllib.urlencode can't handle unicode, this is a hack to fix it."""
     data_iter = None
     if isdict(data):
-        data_iter = data.iteritems()
+        data_iter = data.items()
     elif islist(data):
         data_iter = data
 
     if data_iter:
         for i, (key, value) in enumerate(data_iter):
-            if isinstance(value, unicode):
+            if isinstance(value, text_type):
                 # try to convert to str
                 try:
                     safe_val = str(value)
@@ -79,10 +84,10 @@ def urlencode_unicode(data):
                 finally:
                     if isdict(data):
                         data[key] = safe_val
-                    elif islist(data):
+                    else:
                         data[i] = (key, safe_val)
 
-    return urllib.urlencode(data)
+    return urlencode(data, doseq=doseq)
 
 
 def _is_leaf(value):
@@ -103,7 +108,7 @@ def to_nested_php_args(data, prefix_key=None):
         data_iter = data if is_root else enumerate(data)
         new_data = [] if is_root else {}
     elif isdict(data):
-        data_iter = data.iteritems()
+        data_iter = data.items()
         new_data = {}
     else:
         raise TypeError('expected dict or list, got {0}'.format(type(data)))
@@ -112,9 +117,9 @@ def to_nested_php_args(data, prefix_key=None):
         def data_set(k, v):
             new_data.append((k, v))
         def data_update(d):
-            for k, v in d.iteritems():
+            for k, v in d.items():
                 new_data.append((k, v))
-    elif isdict(new_data):
+    else:
         def data_set(k, v):
             new_data[k] = v
         data_update = new_data.update
@@ -135,23 +140,31 @@ def prepend_base(base):
     return lambda call: '.'.join((base, call))
 
 
-# TODO check against ABCs instead of hasattr
 def isdict(value):
     """Return true if the value behaves like a dict, false if not."""
-    return hasattr(value, 'keys') and hasattr(value, '__getitem__')
+    return isinstance(value, MutableMapping)
 
 
-# TODO check against ABCs instead of hasattr
 def islist(value):
     """Return true if the value behaves like a list, false if not."""
-    return hasattr(value, 'append') and hasattr(value, '__getitem__')
+    return isinstance(value, MutableSequence)
 
 
 def isstr(value):
     """Return true if the value behaves like a string, false if not."""
-    return isinstance(value, basestring)
+    return isinstance(value, string_types)
 
 
-def signature_position(func, arg_name):
-    """Look at func's signature and return the position of arg_name."""
-    return inspect.getargspec(func).args.index(arg_name)
+def get_filename(disposition):
+    """Parse Content-Disposition header to pull out the filename bit.
+
+    See: http://tools.ietf.org/html/rfc2616#section-19.5.1
+
+    """
+    if disposition:
+        params = [param.strip() for param in disposition.split(';')[1:]]
+        for param in params:
+            if '=' in param:
+                name, value = param.split('=', 1)
+                if name == 'filename':
+                    return value.strip('"')
